/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	CLEMENT					*/
/*	Its purpose is to handle SMTP protocole	*/
/*	on SMTP server premises and crunch data	*/
/*	and get clearance from the main Mentor	*/
/*	server to pass the E-mail to local	*/
/*	smtp server (sendmail, postfixe, etc)	*/
/*						*/
/************************************************/
#include	<syslog.h>
#include	<string.h>
#include	<errno.h>
#include	<signal.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	"modtra.h"
#include	"modchk.h"
#include	"modsnd.h"
#include	"modtrp.h"
#include	"lvleml.h"
#include	"lvlspm.h"
#include	"lvluid.h"
#include	"gescnf.h"
#include	"gesclm.h"
#include	"gesipc.h"
#include	"unidev.h"
#include	"unilck.h"
#include	"unieml.h"
#include	"unidns.h"
#include	"uninet.h"
#include	"subrou.h"
#include	"dbgmem.h"

typedef	struct	{
	BOXTYP *boxe;	/*IPC boxe exchange		*/
	char *smtpsrv;	/*local smtp sender		*/
	char *spmdsrv;	/*SPAM detector daemon		*/
	char *rmponder;	/*Mentor local transponder	*/
	pid_t tpdrpid;	/*local transponder pid		*/
	pid_t sndpid;	/*local sender pid		*/
	pid_t chkrpid;	/*local checker pid		*/
	char *chkidsrv;	/*User check-id Server		*/
	char *chkidmod;	/*check-id check mode		*/
	char **frelais;	/*relay list filename		*/
	time_t lastload;/*last mng load			*/
	IPTYP **relais;	/*accepted relay list		*/
	LSTTYP **look;	/*binded channel		*/
	AUXTYP **auxs;	/*auxiliary list		*/
	}MNGTYP;

typedef struct {	/*adv dispatcher struct		*/
	char *checker;	/*checker name			*/
	pid_t pid;	/*sending channel		*/
	char *msg;	/*message to send		*/
	}DISTYP;

static char curpath[300];
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,rou_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-f ] "
		      "[-h ] "
		      "[-v ] "
		      "\n",appname);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-f           : foreground\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to handle one communication	*/
/*	channel with a remote client. return	*/
/*	the foked process pid.			*/
/*						*/
/************************************************/
pid_t action(TRATYP *track,LSTTYP *channel,int num)

{
pid_t child;

switch((child=rou_fork(false))) { 
  case  -1	:
    (void) alert(0,"%s:action Unable to start action process "
                   "(error=<%s>)",appname,strerror(errno));
    break;
  case   0	:	
      {
      int proceed;
      u_int cycle;
  
      proceed=true;
      cycle=0;
      while (proceed==true) {
        (void) settitle("%s waiting on [%s:%d] (%02d/%02d/%04d)",
		         appname,channel->ipnum,
		         channel->port,num+1,channel->iteration,cycle);
        cycle+=tra_tracking(track,channel);
        if ((quickout==false)&&(sigreload==false)) {
	  TIMESPEC timeout;

          timeout.tv_sec=0;
          timeout.tv_nsec=50000;
          while (ipc_waitpost(track->boxe,&timeout)==true) {
	    char *caller;
	    char *rqst;

	    if ((rqst=ipc_getpost(track->boxe,&caller))!=(char *)0) {
	      (void) alert(0,"action, Droping caller->msg=<%s->%s>",
			      caller,rqst);
	      caller=rou_freestr(caller);
  	      rqst=rou_freestr(rqst);
	      }
	    }
	  /*to make sure to have a new uniqID next time			*/
	  (void) nanosleep(&timeout,&timeout);
	  }
	else
	  proceed=false;
        }
      /*Call to check any memory leak if memleak set to 'true', 	*/
      /*request done in debug mode only.				*/
      (void) dbg_dumpmem(appname);
      }
    (void) exit(0);
    break;
  default	:
    break;
  }
return child;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to load the config definition	*/
/*	and return a data structure		*/
/*						*/
/************************************************/
static MNGTYP *prepmng(MNGTYP *mng,BOXTYP *boxe,TRATYP *track)

{
CNFTYP *cnf;

if ((cnf=cnf_getconfig())!=(CNFTYP *)0) {
  int i;

  if (mng==(MNGTYP *)0) {
    mng=(MNGTYP *)calloc(1,sizeof(MNGTYP));
    mng->boxe=boxe;
    }
  mng->look=dev_getlst(mng->look,p_msmtp,cnf->bindlst);
  for (i=0;i<3;i++) {
    char **mout;
    char **minp;

    mout=(char **)0;
    minp=(char **)0;
    switch (i) {
      case 0	:
        mout=&(mng->chkidmod);
	minp=&(cnf->chkidmod);
	break;
      case 1	:
        mout=&(mng->rmponder);
	minp=&(cnf->rmponder);
	break;
      case 2	:
        mout=&(mng->spmdsrv);
	minp=&(cnf->spmdsrv);
	break;
      }
    *mout=rou_freestr(*mout); 
    if (*minp!=(char *)0)
      *mout=strdup(*minp);
    }
  (void) dev_mullisten(mng->look);
  mng->frelais=cnf->frelais;
  cnf->frelais=(char **)0;
  mng->relais=cnf->relais;
  cnf->relais=(IPTYP **)0;
  mng->auxs=cnf->auxs;
  cnf->auxs=(AUXTYP **)0;
  mng->lastload=cnf->lastload;
  track->chkidsrv=strdup(cnf->chkidsrv);
  track->chkidmod=strdup(cnf->chkidmod);
  cnf=cnf_freeconfig(cnf);
  }
return mng;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to start process to listen	*/
/*	on ALL listed interface			*/
/*						*/
/************************************************/
static void dolistening(TRATYP *track,LSTTYP **look)

{
if (look!=(LSTTYP **)0) {
  int i;

  (void) dev_mullisten(look);
  for (i=0;look[i]!=(LSTTYP *)0;i++) {
    int j;

    if (look[i]->handle<0)
      continue;
    for (j=0;j<look[i]->iteration;j++) {
      if (look[i]->process[j]==(pid_t)0) {
        look[i]->process[j]=action(track,look[i],j);
        continue;
        }
      if (kill(look[i]->process[j],0)<0)
        look[i]->process[j]=(pid_t)0;
      }
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to split an authentication	*/
/*	message in 3 components.		*/
/*	username,password,uniqid.		*/
/*						*/
/************************************************/
static int splitmsg(char *msg,char **username,char **password,char **uniqid)

{
int good;
char *ptr;

good=false;
*username=(char *)calloc(strlen(msg),sizeof(char));
*password=(char *)calloc(strlen(msg),sizeof(char));
*uniqid=(char *)calloc(strlen(msg),sizeof(char));
if ((ptr=strchr(msg,':'))!=(char *)0) {
  ptr++;
  (void) strcpy(msg,ptr);
  if (sscanf(msg,"%s %s %s",*username,*password,*uniqid)==3) {
    good=true;
    if ((ptr=strchr(*username,'<'))!=(char *)0) {
      good=false;
      (void) strcpy(*username,ptr+1);
      if ((ptr=strrchr(*username,'>'))!=(char *)0) {
        *ptr='\000';
        good=true;
        }
      }
    }
  }
return good;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to send the temporary uniqid	*/
/*	to the checker to be stored within the	*/
/*	data base, if the check can't do it	*/
/*	message is not sent at all.		*/
/*						*/
/************************************************/
static void sendmyuniqid(MNGTYP *mng,char *username,char *uniqid,char **answ)

{
#define OPUN	"clement.c:sendmyuniqid,"
#define	MXTIME	30	/*waiting 30 seconds max*/


if (mng->chkrpid!=(pid_t)0) {
  char *locmsg;
  char *msgref;

  locmsg=rou_memstrcat((char *)0,"%s %s %s",UNIQID,username,uniqid);
  if ((msgref=ipc_postmsg(mng->boxe,mng->chkrpid,locmsg))!=(char *)0) {
    char *got;
    time_t start;

    got=(char *)0;
    start=rou_systime();
    while (got==(char *)0) {
      (void) usleep(10000);
      if ((got=ipc_getreply(mng->boxe,msgref))==(char *)0) {
	if ((start+MXTIME)<rou_systime()) {
	  break;
	  }
	}
      }
    if (got==(char *)0) {
      (void) alert(0,"%s no answer transmiting uniqid for <%s>",
		      OPUN,username);
      *answ=rou_memstrcat(*answ,"%d Mentor Timeout",CMDOK);
      }
    else {
      *answ=got;
      }
    msgref=rou_freestr(msgref);
    }
  locmsg=rou_freestr(locmsg);
  }
else {
  /*Address provided is not within a know domain, no need to	*/
  /*to send the collected passwd.				*/
  *answ=rou_memstrcat(*answ,"%d",CMDOK);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to send an CKRCPT message to	*/
/*	the proper checker, if the domain is not*/
/*	in the list of known domain but the	*/
/*	origin is within the relayed number, the*/
/*	message is accepted.			*/
/*						*/
/************************************************/
static void dispatchchk(MNGTYP *mng,char *caller,char *msg)

{
char *answ;
int relayed;
RCPTMSG *rm;
char *domain;
int phase;
int proceed;

answ=(char *)0;
relayed=false;
domain=(char *)0;
rm=(RCPTMSG *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case   0	:	/*checking message format	*/
      if (((rm=eml_getrcptmsg(msg))==(RCPTMSG *)0)||
	  (strrchr(rm->rcpt,'@')==(char *)0)) {
        answ=eml_ascstat(BADPAR,msg);
        phase=999;
	}
      break;
    case   1	:	/*refreshing the relay list	*/
      mng->relais=net_mkiplst(mng->frelais,mng->relais,mng->lastload);
      rm->relais|=net_isiniplst(mng->relais,rm->iprem);
      break;
    case   2	:	/*searching for checker		*/
      domain=eml_strdomain(rm->rcpt);
      if (mng->chkrpid!=(pid_t)0) {
	char *locmsg;

	locmsg=rou_memstrcat((char *)0,"%s %s %s %lu %s,%d %s %d",
				        CKRCPT,rm->iprem,rm->from,
					rm->taille,
					rm->listenip,rm->listenport,
					rm->rcpt,rm->relais);
	(void) ipc_relaymsg(mng->boxe,caller,mng->chkrpid,locmsg);
	locmsg=rou_freestr(locmsg);
	proceed=false;	/*no need to go further		*/
	}
      break;
    case   3	:	/*to bad, no checker process	*/
      answ=eml_ascstat(ERRPROC,rm->rcpt);
      break;
    default	:	/*reporting trouble to caller	*/
      if (answ!=(char *)0) {
        (void) ipc_postreply(mng->boxe,caller,answ);
        /*answ is a static memory			*/
        }
      proceed=false;
      break;
    }
  phase++;
  }
rm=eml_freercptmsg(rm);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to send an ADVICE/SENT	*/
/*	message to checker process if available.*/
/*						*/
/************************************************/
static void dispatchecker(MNGTYP *mng,char *caller,char *mtype,char *msg)

{
int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*call for Mentor advice*/
      if (mng->chkrpid!=(pid_t)0) {
	char *locmsg;

	locmsg=rou_memstrcat((char *)0,"%s %s",mtype,msg);
	(void) ipc_relaymsg(mng->boxe,caller,mng->chkrpid,locmsg);
	locmsg=rou_freestr(locmsg);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 1	:	/*no Mentor Available	*/
      (void) alert(0,"%s, %s msg not forwarded, checker missing",
		     "clement.c:dispatchecker",mtype);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check user ID using PAM	*/
/*	or POP.					*/
/*						*/
/************************************************/
static void docheckid(MNGTYP *mng,char *caller,char *msg)

{
int chkstat;
char *answ;

chkstat=ERRPROC;
answ=(char *)0;
switch(rou_fork(false)) { 
  case  -1	:
    (void) alert(0,"%s,loc_emailmgs, Unable to start action process "
                   "(error=<%s>)",appname,strerror(errno));
    (void) ipc_postreply(mng->boxe,caller,eml_ascstat(chkstat,answ));
    break;
  case   0	: {
    char *username;
    char *password;
    char *uniqid; 

    username=(char *)0;
    password=(char *)0;
    uniqid=(char *)0;
    if (splitmsg(msg,&username,&password,&uniqid)==true) {
      switch (chkstat=uid_chkid(mng->chkidmod,username,password)) {
        case CMDOK	:
	  if (strcasecmp(uniqid,"void")!=0)
            (void) sendmyuniqid(mng,username,uniqid,&answ);
	  else	/*SMTP authentication	*/
	    {
            answ=strdup(eml_ascstat(IDOK));
	    }
	  break;
        default		:
          answ=strdup(eml_ascstat(BADID,"Authentication failure"));
	  break;
        }
      }
    else {
      answ=strdup(eml_ascstat(NOOP,answ));
      }
    (void) ipc_postreply(mng->boxe,caller,answ);
    answ=rou_freestr(answ);
    uniqid=rou_freestr(uniqid);
    password=rou_freestr(password);
    username=rou_freestr(username);
    (void) exit(0);
    }
    break;
  default	:	
    (void) usleep(1000);
    break;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to do received messages	*/
/*	screening and relay messages to the	*/
/*	pertaining checker.			*/
/*						*/
/************************************************/
static void dispatchmsg(MNGTYP *mng,char *caller,char *msg)

{
#define	OPDI	"clement.c:dispatchmsg,"


typedef	enum	{
	d_ckrcpt,	/*rcpt request		*/
	d_advice,	/*advice request	*/
	d_chkid,	/*checkid request	*/
	d_sent,		/*sent Email information*/
	d_unknown	/*unknow request	*/
	}DENUM;

static VOCTYP vocdis[]={
        {d_ckrcpt,CKRCPT,(const void *)0},
        {d_advice,ADVICE,(const void *)0},
        {d_chkid,CHKID,(const void *)0},
        {d_sent,SENT,(const void *)0},
        {d_unknown,(char *)0,(const void *)0}
        };

VOCTYP *voc;

if ((voc=rou_getvoca(vocdis,msg))!=(VOCTYP *)0) {
  switch (voc->code) {
    case d_ckrcpt	:
      (void) dispatchchk(mng,caller,msg);
      break;
    case d_advice	:
      (void) dispatchecker(mng,caller,ADVICE,msg);
      break;
    case d_sent	:
      (void) dispatchecker(mng,caller,SENT,msg);
      break;
    case d_chkid	:
      (void) docheckid(mng,caller,msg);
      break;
    default		:
      (void) alert(0,"%s Unknown message <%s>",OPDI,msg);
      (void) ipc_postreply(mng->boxe,caller,
			   eml_ascstat(CMDBAD,"Unknown command to dispatch"));
      break;
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check child process and	*/
/*	make sure they are up and running.	*/
/*						*/
/************************************************/
static void checkchild(MNGTYP *mng)

{
#define OPDR	"clement.c:checkchild,"

static time_t lastcheck=(time_t)0;

time_t curtime;

curtime=rou_systime();
if ((lastcheck+5)<=curtime) {
  int i;

  for (i=0;i<3;i++) {
    pid_t child;
    char *alrt;

    child=(pid_t)0;
    alrt="";
    switch(i) {
      case   0	:
	child=mng->sndpid;
        alrt="sender";
	break;
      case   1	:
	child=mng->chkrpid;
        alrt="checker";
	break;
      case   2	:
        if (mng->rmponder!=(char *)0)
	  child=mng->tpdrpid;
        alrt="transponder";
	break;
      }
    if (rou_checkprocess(child)==false) {
      if (lastcheck!=(time_t)0)
        (void) alert(0,"%s detect '%s' interface missing",OPDR,alrt);
      switch (i) {
	case 0	:
  	  mng->sndpid=snd_manager(mng->boxe,mng->auxs);
	  break;
	case 1	:
          mng->chkrpid=chk_manager(mng->boxe,mng->auxs);
	  break;
	case 2	:
	  if (mng->rmponder!=(char *)0)
            mng->tpdrpid=trp_manager(mng->rmponder);
	  break;
        }
      }
    }
  lastcheck=curtime;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to stop all managed process 	*/
/*	and free the data structure.		*/
/*						*/
/************************************************/
static MNGTYP *dropmng(MNGTYP *mng)

{
if (mng!=(MNGTYP *)0) {
  pid_t ksend[4];

  (void) dev_mulclose(mng->look);
  mng->look=dev_freelst(mng->look);
  ksend[0]=mng->sndpid;
  ksend[1]=mng->chkrpid;
  ksend[2]=mng->tpdrpid;
  (void) rou_childkill(ksend,3,5);
  mng->rmponder=rou_freestr(mng->rmponder);
  mng->frelais=rou_freestrlst(mng->frelais);
  mng->spmdsrv=rou_freestr(mng->spmdsrv);
  mng->chkidmod=rou_freestr(mng->chkidmod);
  (void) free(mng);
  mng=(MNGTYP *)0;
  }
return mng;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to wait (briefly) for a 	*/
/*	message comming from listining interface*/
/*	and dispatch message accordingly	*/
/*	handle requested comming from SMTP	*/
/*	Stay (at most) 2 sec in this procedure	*/
/*						*/
/************************************************/
static void procmsg(MNGTYP *mng)

{
#define	WSEC	2

TIMESPEC timeout;

timeout.tv_nsec=0;
timeout.tv_sec=WSEC;
while (ipc_waitpost(mng->boxe,&timeout)>0) {
  char *caller;
  char *msg;

  if ((msg=ipc_getpost(mng->boxe,&caller))!=(char *)0) {
    (void) dispatchmsg(mng,caller,msg);
    (void) free(msg);
    caller=rou_freestr(caller);
    }
  if (quickout==true)
    break;
  }
(void) ipc_cleanbox(mng->boxe);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to open binding channel and	*/
/*	handle requested comming from SMTP	*/
/*	client.					*/
/*						*/
/************************************************/
static void  cycle()

{
#define	MXBOX	1000	/*Max 1000 messages queued*/

BOXTYP *boxe;
MNGTYP *mng;
TRATYP *track;

boxe=ipc_getbox(MXBOX);
mng=(MNGTYP *)0;
track=(TRATYP *)calloc(1,sizeof(TRATYP));
track->boxe=boxe;
track->clam=clm_opentree(track->clam);
(void) settitle("%s main daemon -d%d -b%s",appname,debug,base);
while (quickout==false) {
  if ((mng=prepmng(mng,boxe,track))==(MNGTYP *)0) {
    (void) alert(0,"%s configuration file missing altogether (aborting)",
		    appname);
    break;
    }
  (void) spm_isspamdup(mng->spmdsrv);
  (void) dolistening(track,mng->look);
  (void) checkchild(mng);
  (void) procmsg(mng);
  }
mng=dropmng(mng);
track->clam=clm_closetree(track->clam);
(void) free(track);
boxe=ipc_dropbox(boxe);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
char c;
STDDBG *dbglog;

base=strdup(base);
(void) strncpy(curpath,argv[0],sizeof(curpath));
(void) dbg_setcurpath(argv[0]);
(void) openlog(appname,LOG_PID,0);
while ((c=getopt(argc,argv,"b:d:fhv"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=rou_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'f'  :
      foreground=true;
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
if ((debug==0)||(verbose==true)) {
  char curvers[100];
  char msg[200];

  (void) snprintf(curvers,sizeof(curvers),"%s",rou_getvers());
  (void) strcpy(msg,"");
  switch (dns_chksafeversion(curvers,sizeof(curvers))) {
    case -1	:
       (void) snprintf(msg,sizeof(msg),"\nYour current %s version is <%s> "
				       "while the last released one is <%s>.\n"
				        "You should update your %s program",
				        appname,rou_getvers(),curvers,appname);
      break;
    case  0	:
      break;
    case +1	:
       (void) snprintf(msg,sizeof(msg),"\nYour current %s-%s is a BETA "
				       "version.\n"
				       "Please report any discrepancy or bugs "
				       "you may find to <clement@safe.ca>.\n",
				        appname,rou_getvers());
      break;
    }
  if (strlen(msg)>0) {
    (void) fprintf(stderr,"%s\n",msg);
    }
  }
(void) rou_chgid(appname);
if ((strlen(base)>0)&&(chdir(base)<0)) {
  (void) fprintf(stderr,"Unable to reach the base directory <%s>\n",base);
  return -1;
  }
(void) divedivedive(foreground);
if (lck_locking(appname,LCK_LOCK,5)==true) {
  dbglog=eml_switchstd(appname);
  (void) setjournal(APJRL,d_log,appname);
  (void) journalling(true,APJRL,"%s-%s started",appname,rou_getvers());
  (void) setjournal(EVJRL,d_log,"events");
  (void) preptitle(argc,argv,environ);
  (void) rou_setalrm();
  (void) cycle();
  (void) cleantitle();
  (void) journalling(true,APJRL,"%s-%s stopped",appname,rou_getvers());
  dbglog=eml_closeswitchstd(dbglog);
  (void) closejournals();
  (void) lck_locking(appname,LCK_UNLOCK,1);
  }
(void) closelog();
(void) free(base);
return 0;
}
