/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESCLM:					*/
/*	Take Care of all "CLAMAV"  anti-virus	*/
/*	management.				*/
/*						*/
/************************************************/
#include	<string.h>
#include	<setjmp.h>
#include	<errno.h>
#include	<signal.h>
#include	<syslog.h>
#include	<stdlib.h>
#include        <execinfo.h>
#include	<clamav.h>
#include	"subrou.h"
#include	"unidns.h"
#include	"gesclm.h"

typedef	struct	{
	time_t last;		/*last load	*/
	int retry;		/*retry time;	*/
	struct cl_node *virlst;	/*virus list	*/
	struct cl_stat dbstat;	/*dir status	*/
	char infos[300];	/*virus infos	*/
	}CLAMFILE;

#define	CLMCONF		"freshclam.conf"
#define	CLMDIRINFO	"DatabaseDirectory"
#define	CLMVINFO	"DNSDatabaseInfo"

#define CLMREFRESH	3600	/*every hour	*/

static jmp_buf trouble;		/*clamav crashed*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to load clamav variable	*/
/*	version					*/
/*						*/
/************************************************/
static void loadclamenv()

{
FILE *fichier;
char filename[400];

(void) setenv(CLMVINFO,"current.cvd.clamav.net",1);
(void) setenv(CLMDIRINFO,"/var/lib/clamav",1);
(void) snprintf(filename,sizeof(filename),"/%s/%s","etc",CLMCONF);
if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  char line[500];

  while (rou_getstr(fichier,line,sizeof(line),'#')!=(char *)0) {
    char var[100];
    char contents[200];

    if (strlen(line)==0)	/*empty line	*/
      continue;
    (void) memset(var,'\000',sizeof(var));
    (void) memset(contents,'\000',sizeof(contents));
    if (sscanf(line,"%99s %199s",var,contents)==2) {
      if (strcmp(var,CLMVINFO)==0) {
	(void) setenv(CLMVINFO,contents,1);
	}
      if (strcmp(var,CLMDIRINFO)==0) {
	(void) setenv(CLMDIRINFO,contents,1);
	}
      }
    }
  (void) fclose(fichier);
  }
else
  (void) alert(1,"gesclm.c:getlastvers,  unable to open file '%s' "
		 "(error=<%s>), guessing values",filename,strerror(errno));
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the clamav data-base*/
/*	directory.				*/
/*						*/
/************************************************/
static char *retdbdir()

{
static char *ptrdb=(char *)0;
static char dbdir[300];

if (ptrdb==(char *)0) {
  (void) snprintf(dbdir,sizeof(dbdir),"%s",getenv(CLMDIRINFO));
  ptrdb=dbdir;
  }
return ptrdb;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch CLAMV SEGV  signal.	*/
/*	(big trouble deep insides clamv)	*/
/*	we fork the process and then coredump	*/
/*	the copy such investigation can proceed	*/
/*	later one.				*/
/*						*/
/************************************************/
static void bigpbs(int sig)

{
switch (fork()) {
  case -1	:
    (void) alert(0,"gesclm.c:bigbps, unable to fork (error=<%s>)",
		    strerror(errno));
    break;
  case  0	:
    (void) core_dump("Crash of clamav, Event core_dump in progress");
    break;
  default	:
    (void) siglongjmp(trouble,1);
    break;
  }
(void) alert(0,"gesclm.c:bigbps, unexpected exit under clam crash (bug!)");
(void) exit(-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load the virus directory	*/
/*						*/
/************************************************/
int loadclamdir(CLAMFILE *virus)

{
int status;

if (virus->virlst!=(struct cl_node *)0) {
  (void) cl_free(virus->virlst);
  virus->virlst=(struct cl_node *)0;
  }
if ((status=cl_loaddbdir(retdbdir(),&(virus->virlst),(unsigned int *)0))==0) {
  time_t curtime;
  char current[100];
  

  curtime=rou_systime();
  (void) cl_build(virus->virlst);
  (void) cl_statfree(&(virus->dbstat));
  (void) cl_statinidir(retdbdir(),&(virus->dbstat));
  
  (void) snprintf(virus->infos,sizeof(virus->infos),
		  "Clamav-%s (virlib:%s EST5EDT)",
		  cl_retver(),rou_getstrdatetime(curtime));
  (void) snprintf(current,sizeof(current),"%s",cl_retver());
  if (dns_chkversion(getenv(CLMVINFO),current,sizeof(current))!=true) {
    (void) alert(0,"Please install Clamav-%s",current);
    (void) journalling(false,APJRL,"Please install Clamav-%s",current);
    }
  (void) journalling(true,APJRL,"Clamav-%s virus library reloaded "
			        "(process=%05d)",cl_retver(),getpid());
  virus->last=curtime;
  virus->retry=CLMREFRESH;	/*one hour	*/
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if a new virus tree is	*/
/*	needed to be loaded			*/
/*						*/
/************************************************/
CLAMTYP *clm_refreshtree(CLAMTYP *clamav)

{
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *virus;
  time_t curtime;

  virus=(CLAMFILE *)clamav;
  curtime=rou_systime();
  if ((virus->last+virus->retry)<curtime) {
    if (cl_statchkdir(&(virus->dbstat))==1) {
      if (virus->virlst!=(struct cl_node *)0) {
        (void) cl_free(virus->virlst);
        virus->virlst=(struct cl_node *)0;
        }
      (void) cl_statfree(&(virus->dbstat));
      if ((loadclamdir(virus),(int *)0)!=0) {
        virus->retry=600;	/*ten minutes	*/
        }
      }
    virus->last=curtime;
    }
  }
return clamav;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a virus scan list	*/
/*						*/
/************************************************/
CLAMTYP *clm_opentree(CLAMTYP *clamav)

{
#define	OPTREE	"gesclm.c:clm_opentree,"
#define	TMPDIR	"TMPDIR"

if (clamav==(CLAMTYP *)0) {
  (void) loadclamenv();
  clamav=(CLAMFILE *)calloc(1,sizeof(CLAMFILE));
  if (loadclamdir(clamav)!=0) {
    (void) free(clamav);
    clamav=(CLAMFILE *)0;
    (void) alert(0,"%s Unable to open Virus tree (bug?)",OPTREE);
    }
  }
else {
  clamav=clm_refreshtree(clamav);
  }
return (CLAMTYP *)clamav;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free a virus scan list	*/
/*						*/
/************************************************/
CLAMTYP *clm_closetree(CLAMTYP *clamav)

{
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *virus;

  virus=(CLAMFILE *)clamav;
  (void) cl_statfree(&(virus->dbstat));
  (void) cl_free(virus->virlst);
  (void) free(virus);
  virus=(CLAMFILE *)0;
  clamav=(CLAMTYP *)virus;
  }
return clamav;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if a virus is found	*/
/*	within the E-mail contents.		*/
/*	If visur discovered, return the virus	*/
/*	name.					*/
/*						*/
/************************************************/
CLMCHKTYP clm_checkvirus(CLAMTYP *clamav,char *filename,const char **virname)

{
#define	MXCLM	100

CLMCHKTYP status;
int result;

status=clm_nocheck;
*virname=(char *)0;
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *virus;
  struct cl_limits limits;
  void (*prvsegv)(int);


  virus=(CLAMFILE *)clamav;
  (void) memset(&limits,'\000',sizeof(struct cl_limits));
  limits.maxreclevel=10;
  limits.maxfiles=1000;
  limits.maxfilesize=MXCLM*(1024*1024);
  limits.maxreclevel=10;
  prvsegv=signal(SIGSEGV,bigpbs);
  if (sigsetjmp(trouble,true)==0) {
    result=cl_scanfile(filename,virname,(unsigned long int *)0,
			        virus->virlst,&limits,
			        CL_SCAN_STDOPT);
    switch (result) {
      case CL_CLEAN	:
        status=clm_clean;
        break;
      case CL_VIRUS	:
        status=clm_virus;
        break;
      default		:
        (void) journalling(false,APJRL,"Virus scanning error <%s> "
				       "found on <%s>",cl_perror(result),
					strrchr(filename,'/')+1);
        (void) tracing(TRALOG,0,"gesclm.c:clm_checknotvirus, error=<%s>",
				 cl_perror(result));
        break;
      }
    }
  else {
    (void) journalling(false,APJRL,"Virus scanning crashed on <%s>",filename);
    }
  (void) signal(SIGSEGV,prvsegv);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to return the clamd version	*/
/*	and the last update date.		*/
/*						*/
/************************************************/
const char *clm_getversdate(CLAMTYP *clamav)

{
CLAMFILE *virus;

virus=(CLAMFILE *)clamav;
return (const char *)virus->infos;
}
