/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESCNF:					*/
/*	Take Care of all config definition	*/
/*						*/
/************************************************/
#include	<malloc.h>
#include	<syslog.h>
#include	<stdio.h>
#include	<string.h>
#include	<unistd.h>
#include	<ctype.h>
#include	<errno.h>
#include	"subrou.h"
#include	"unieml.h"
#include	"uninet.h"
#include	"gescnf.h"

/*smtp access to local mailer			*/
#define	LOCMAIL	"127.0.0.25,25"


#define	CNF	".conf"	/*config extension	*/

/*config file vocabulary			*/
#define	ADMIN		"ADMIN"
#define	SPMDSRV		"SPMDSRV"
#define	SMTPSRV		"SMTPSRV"
#define	RMPONDER	"RMPONDER"
#define	DOMAIN		"DOMAIN"
#define	BINDING		"BINDING"
#define	RELAY		"RELAY"
#define	CHECKER		"CHECKER"
#define	SENDER		"SENDER"
#define	CHKIDSRV	"CHKIDSRV"
#define	CHKIDMOD	"CHKIDMOD"
#define	DFLTCODE	"DFLTCODE"
#define	CHKCODE		"CHKCODE"
/*

*/
/************************************************/
/*						*/
/*	Procedure to split relai information	*/
/*						*/
/************************************************/
static char **splitlist(char *info)

{
char **list;
char *myinfo;

list=(char **)0;
myinfo=(char *)0;
if (info!=(char *)0)
  myinfo=strdup(info);
info=myinfo;
while ((info!=(char *)0)&&(*info!='\000')) {
  char *ptr;

  if ((ptr=strchr(info,':'))!=(char *)0) {
    *ptr='\000';
    ptr++;
    }
  if (strlen(info)>0)
    list=(char **)rou_addveclst((void **)list,(void *)strdup(info));
  info=ptr;
  }
myinfo=rou_freestr(myinfo);
return list;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to split domain infomation	*/
/*	according domain separator.		*/
/*						*/
/************************************************/
static char **mulsplitlist(char **list,char *doms)

{
char **nlist;

if ((nlist=splitlist(doms))!=(char **)0) {
  int i;

  for (i=0;nlist[i]!=(char *)0;i++) {
    list=(char **)rou_addveclst((void **)list,(void *)nlist[i]);
    }
  (void) free(nlist);
  }
return list;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a default auxiliary	*/
/*	to do relaying.				*/
/*						*/
/************************************************/
static void addrelaying(CNFTYP *cnf)

{
AUXTYP *auxs;
char line[100];

auxs=(AUXTYP *)calloc(1,sizeof(AUXTYP));
(void) snprintf(line,sizeof(line),"%s,25",LOCALHOST);
auxs->sender=strdup(line);
auxs->relay=true;
auxs->domains=(char **)rou_addveclst((void **)0,strdup(RLAYDOM));
cnf->auxs=(AUXTYP **)rou_addveclst((void **)cnf->auxs,(void *)auxs);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return an SKR location 	*/
/*	if exsiting.				*/
/*						*/
/************************************************/
static SKRTYP *findskr(SKRTYP **lst,CKPREF pref,char *hostname,char *chkcode)

{
SKRTYP *sk;

sk=(SKRTYP *)0;
if (lst!=(SKRTYP **)0) {
  int i;

  for (i=0;lst[i]!=(SKRTYP *)0;i++) {
    int found;

    found=false;
    if (lst[i]->pref!=pref)
      continue;
    if (strcmp(lst[i]->hostname,hostname)!=0)
      continue;
    if (pref==p_sender)
      found=true;
    else {
      if ((chkcode==(char *)0)&&(lst[i]->chkcode)==(char *)0)
        found=true;
      else {
        if ((chkcode!=(char *)0)&&(lst[i]->chkcode)!=(char *)0) 
          if (strcmp(lst[i]->chkcode,chkcode)==0) 
	    found=true;
        }
      }
    if (found==true) {
      sk=lst[i];
      break;
      }
    } 
  }
return sk;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a sndr/chkr to a list	*/
/*						*/
/************************************************/
static SKRTYP **addskr(SKRTYP **lst,CKPREF pref,char *hostname,AUXTYP *aux)

{
if (hostname!=(char *)0) {
  SKRTYP *sk;

  if ((sk=findskr(lst,pref,hostname,aux->chkcode))==(SKRTYP *)0) {
    sk=(SKRTYP *)calloc(1,sizeof(SKRTYP));
    sk->pref=pref;
    sk->hostname=strdup(hostname); 
    sk->relay=aux->relay;
    if (aux->chkcode!=(char *)0)
      sk->chkcode=strdup(aux->chkcode); 
    lst=(SKRTYP **)rou_addveclst((void **)lst,(void *)sk);
    }
  if (aux->domains!=(char **)0) {
    int i;
  
    for (i=0;aux->domains[i]!=(char *)0;i++) {
      sk->doms=(char **)rou_addveclst((void **)sk->doms,
				      (void *)strdup(aux->domains[i]));
      }
    }
  if (aux->bindlst!=(char **)0) {
    int i;
  
    for (i=0;aux->bindlst[i]!=(char *)0;i++) {
      sk->bindlst=(char **)rou_addveclst((void **)sk->bindlst,
				         (void *)strdup(aux->bindlst[i]));
      }
    }
  }
return lst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load the clement conf file	*/
/*	and extract domain, binding and checker	*/
/*	informations.				*/
/*						*/
/************************************************/
CNFTYP *cnf_readconfig()

{
#define	OPCN	"gescnf.c:getconfig,"

#define	DCHKIDSRV	"127.0.0.25,411,1"

typedef	enum	{
	c_admin,	/*clement admin			*/
	c_spmdsrv,	/*Spam daemon definition	*/	
	c_smtpsrv,	/*Local default SMTP srv	*/
	c_rmponder,	/*Mentor local transponder	*/
	c_domain,	/*Domain definition		*/
	c_binding,	/*Binding interface def		*/
	c_relay,	/*Relay accepted def		*/
	c_checker,	/*checker server def		*/
	c_sender,	/*local SMTP sender def		*/	
	c_chkidsrv,	/*local checkid server		*/
	c_chkidmod,	/*local checkid mode		*/
	c_chkcode,	/*checker access code		*/
	c_dfltcode,	/*chkr dflt access code		*/
	c_unknown	/*Unknown keyword		*/
	}ENUCNF;

static VOCTYP vocable[]={
		{c_admin,ADMIN,(const void *)0},
		{c_spmdsrv,SPMDSRV,(const void *)0},
		{c_smtpsrv,SMTPSRV,(const void *)0},
		{c_rmponder,RMPONDER,(const void *)0},
		{c_domain,DOMAIN,(const void *)0},
		{c_binding,BINDING,(const void *)0},
		{c_relay,RELAY,(const void *)0},
		{c_checker,CHECKER,(const void *)0},
		{c_sender,SENDER,(const void *)0},
		{c_chkidsrv,CHKIDSRV,(const void *)0},
		{c_chkidmod,CHKIDMOD,(const void *)0},
		{c_dfltcode,DFLTCODE,(const void *)0},
		{c_chkcode,CHKCODE,(const void *)0},
		{c_unknown,(char *)0,(const void *)0}
		};

CNFTYP *cnf;
FILE *fichier;
char fname[400];

cnf=(CNFTYP *)0;
(void) rou_mkname(fname,sizeof(fname),d_etc,"",appname,CNF);
if ((fichier=fopen(fname,"r"))!=(FILE *)0) {
  AUXTYP *auxs;
  char line[3000];

  auxs=(AUXTYP *)0;
  cnf=(CNFTYP *)calloc(1,sizeof(CNFTYP));
  cnf->chkidsrv=strdup(DCHKIDSRV);
  cnf->dfltcode=strdup("Unassigned");
  cnf->chkidmod=strdup("PAM");
  cnf->smtpsrv=strdup(LOCMAIL);
  while (rou_getstr(fichier,line,sizeof(line),'#')!=(char *)0) {
    VOCTYP *voc;
    char *ptr;

    if ((ptr=strchr(line,'='))!=(char *)0)
      *ptr=' ';
    if ((voc=rou_getvoca(vocable,line))!=(VOCTYP *)0) {
      switch (voc->code) {
        case c_admin	:	/*No Care	*/
	  cnf->adm=(char **)rou_addveclst((void **)cnf->adm,
					  (void *)strdup(line));
	  break;
        case c_spmdsrv	:
	  cnf->spmdsrv=rou_freestr(cnf->spmdsrv);
	  cnf->spmdsrv=strdup(line);
	  break;
        case c_smtpsrv	:
	  cnf->smtpsrv=rou_freestr(cnf->smtpsrv);
	  cnf->smtpsrv=strdup(line);
	  break;
        case c_rmponder	:
	  cnf->rmponder=rou_freestr(cnf->rmponder);
	  cnf->rmponder=strdup(line);
	  break;
        case c_relay	:
	  if (isdigit(line[0])==0) {
	    /*only one relay file name		*/
	    cnf->frelais=rou_freestrlst(cnf->frelais);
	    cnf->frelais=splitlist(line);
	    }
	  else
	    cnf->relais=net_addiplst(cnf->relais,line);
	  break;
        case c_domain	:
	  if (auxs!=(AUXTYP *)0)
	    auxs->domains=mulsplitlist(auxs->domains,line);
	  else
            (void) alert(0,"%s DOMAIN info <%s> is NOT within binding "
			   "definition (file=<%s>)",OPCN,line,fname);
	  break;
        case c_binding	:
	  auxs=(AUXTYP *)calloc(1,sizeof(AUXTYP));
	  auxs->bindlst=splitlist(line);
	  auxs->sender=strdup(cnf->smtpsrv);
	  auxs->chkcode=strdup(cnf->dfltcode);
	  cnf->auxs=(AUXTYP **)rou_addveclst((void **)cnf->auxs,(void *)auxs);
	  if (cnf->bindlst!=(char *)0)
            cnf->bindlst=rou_memstrcat(cnf->bindlst,":");
          cnf->bindlst=rou_memstrcat(cnf->bindlst,"%s",line);
	  break;
        case c_dfltcode	:
	  cnf->dfltcode=rou_freestr(cnf->dfltcode);
	  cnf->dfltcode=strdup(line);
	  break;
        case c_chkcode	:
	  if (auxs!=(AUXTYP *)0) {
	    auxs->chkcode=rou_freestr(auxs->chkcode);
	    auxs->chkcode=strdup(line);
	    }
	  else
            (void) alert(0,"%s CHKCODE info <%s> is NOT within binding "
			   "definition (file=<%s>)",OPCN,line,fname);
	  break;
        case c_checker	:
	  if (auxs!=(AUXTYP *)0) {
	    auxs->checker=rou_freestr(auxs->checker);
	    auxs->checker=strdup(line);
	    }
	  else
            (void) alert(0,"%s CHECKER info <%s> is NOT within binding "
			   "definition (file=<%s>)",OPCN,line,fname);
	  break;
        case c_sender	:
	  if (auxs!=(AUXTYP *)0) {
	    auxs->sender=rou_freestr(auxs->sender);
	    auxs->sender=strdup(line);
	    }
	  else
            (void) alert(0,"%s SENDER info <%s> is NOT within domain "
			   "definition (file=<%s>)",OPCN,line,fname);
	  break;
        case c_chkidsrv	:
	  cnf->chkidsrv=rou_freestr(cnf->chkidsrv);
	  cnf->chkidsrv=strdup(line);
	  break;
        case c_chkidmod	:
	  cnf->chkidmod=rou_freestr(cnf->chkidmod);
	  cnf->chkidmod=strdup(line);
	  break;
	default		:
          (void) alert(0,"%s VOC->%d, (bug?) can't reach keyword in <%s>",
			 OPCN,voc->code,line);
	  break;
	}
      }
    }
  (void) fclose(fichier);
  if ((cnf->frelais==(char **)0)&&(cnf->relais==(IPTYP **)0)) {
    cnf->relais=net_addiplst(cnf->relais,"127.0.0.1/32");
    }
  if (cnf->bindlst!=(char *)0)
    cnf->bindlst=rou_memstrcat(cnf->bindlst,":");
  cnf->bindlst=rou_memstrcat(cnf->bindlst,"%s",cnf->chkidsrv);
  (void) addrelaying(cnf);
  }
return cnf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load the current config,	*/
/*	if the file temporary config file was	*/
/*	changed since last load, the config is	*/
/*	reloaded, otherwise config is kept and 	*/
/*	NULL is returned.			*/
/*						*/
/************************************************/
CNFTYP *cnf_getconfig()

{
static time_t lastload=(time_t)0;
static time_t lastcheck=(time_t)0;

CNFTYP *cnf;
time_t curtime;

cnf=(CNFTYP *)0;
curtime=rou_systime();
if ((lastcheck+10)<curtime) {
  char fname[300];

  (void) rou_mkname(fname,sizeof(fname),d_etc,"",appname,CNF);
  switch (rou_isnewfile(fname,lastload)) {
    case   -1 	:
      (void) alert(0,"gescnf.c:cnf_getconfig, Unable to stat file <%s>",fname);
      break;
    case    0	:
      break;
    case    1	:
      cnf=cnf_readconfig();
      cnf->lastload=curtime;
      lastload=curtime;
      break;
    }
  lastcheck=curtime;
  }
return cnf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by	*/
/*	auxiliary information.			*/
/*						*/
/************************************************/
AUXTYP **cnf_freeauxs(AUXTYP **auxs)

{
if (auxs!=(AUXTYP **)0) {
  int i;

  for (i=0;auxs[i]!=(AUXTYP *)0;i++) {
    AUXTYP *dptr;

    dptr=auxs[i];
    dptr->domains=rou_freestrlst(dptr->domains);
    dptr->checker=rou_freestr(dptr->checker);
    dptr->sender=rou_freestr(dptr->sender);
    dptr->chkcode=rou_freestr(dptr->chkcode);
    dptr->bindlst=rou_freestrlst(dptr->bindlst);
    (void) free(dptr);
    }
  (void) free(auxs);
  auxs=(AUXTYP **)0;
  }
return auxs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by	*/
/*	domain information.			*/
/*						*/
/************************************************/
CNFTYP *cnf_freeconfig(CNFTYP *config)

{
if (config!=(CNFTYP *)0) {
  config->auxs=cnf_freeauxs(config->auxs);
  config->relais=net_freeiplst(config->relais);
  config->frelais=rou_freestrlst(config->frelais);
  config->rmponder=rou_freestr(config->rmponder);
  config->spmdsrv=rou_freestr(config->spmdsrv);
  config->smtpsrv=rou_freestr(config->smtpsrv);
  config->chkidsrv=rou_freestr(config->chkidsrv);
  config->dfltcode=rou_freestr(config->dfltcode);
  config->chkidmod=rou_freestr(config->chkidmod);
  config->adm=rou_freestrlst(config->adm);
  (void) free(config);
  config=(CNFTYP *)0;
  }
return config;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to locate the auxiliary name	*/
/*	related to a specific domain.		*/
/*						*/
/************************************************/
AUXTYP *cnf_getaux(AUXTYP **auxs,char *domain,int sndr)

{
AUXTYP *aux;

aux=(AUXTYP *)0;
if (auxs!=(AUXTYP **)0) {
  int i;

  for (i=0;auxs[i]!=(AUXTYP *)0;i++) {
    if (eml_locatedomain(auxs[i]->domains,domain)==true) {
      char *ptr;

      if (sndr==true)
        ptr=auxs[i]->sender;
      else
        ptr=auxs[i]->checker;
      if (ptr!=(char *)0)
	aux=auxs[i];
      break;
      }
    }
  }
return aux;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to make a merged list of 	*/
/*	sender/checker from the auxiliary list.	*/
/*						*/
/************************************************/
SKRTYP **cnf_getskrlst(AUXTYP **auxs)

{
SKRTYP **lst;

lst=(SKRTYP **)0;
if (auxs!=(AUXTYP **)0) {
  int i;

  for (i=0;auxs[i]!=(AUXTYP *)0;i++) {
    lst=addskr(lst,p_checker,auxs[i]->checker,auxs[i]);
    lst=addskr(lst,p_sender,auxs[i]->sender,auxs[i]);
    }
  }
return lst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by an	*/
/*	skr list				*/
/*						*/
/************************************************/
SKRTYP **cnf_freeskrlst(SKRTYP **skrlst)

{
if (skrlst!=(SKRTYP **)0) {
  int i;

  for (i=0;skrlst[i]!=(SKRTYP *)0;i++) {
    skrlst[i]->bindlst=rou_freestrlst(skrlst[i]->bindlst);
    skrlst[i]->doms=rou_freestrlst(skrlst[i]->doms);
    skrlst[i]->chkcode=rou_freestr(skrlst[i]->chkcode);
    skrlst[i]->hostname=rou_freestr(skrlst[i]->hostname);
    (void) free(skrlst[i]);
    }
  (void) free(skrlst);
  skrlst=(SKRTYP **)0;
  }
return skrlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to locate an SKR reference	*/
/*	among an skr list			*/
/*						*/
/************************************************/
SKRTYP *cnf_getskr(SKRTYP **skrlst,char *hostname)

{
SKRTYP *found;

found=(SKRTYP *)0;
if (skrlst!=(SKRTYP **)0) {
  int i;

  for (i=0;skrlst[i]!=(SKRTYP *)0;i++) {
    if (strcmp(skrlst[i]->hostname,hostname)==0) {
      found=skrlst[i];
      break;
      }
    }
  }
return found;
}
