/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESIPC:					*/
/*	Take care on message exchange between	*/
/*	process.				*/
/*						*/
/************************************************/
#include	<string.h>
#include	<signal.h>
#include	<errno.h>
#include	<syslog.h>
#include	"subrou.h"
#include	"gesipc.h"

#define	MINF	400	/*max uniq msg size	*/

typedef	struct	{
	pid_t pidref;	/*message origine	*/
	u_long ctxref;	/*message context number*/
	}MSGUID;

typedef	struct	{
	time_t stamp;	/*time stamp		*/
	pid_t rcpt;	/*message recipient	*/
	pid_t from;	/*message origine	*/
	pid_t  pidref;	/*original sender ref	*/
	u_long ctxref;	/*message context ref	*/
	u_long msgnum;	/*current msgnum	*/
	char info[MINF];/*message contents	*/
	}MSGINF;

typedef	struct	{
	u_long msgnum;	/*current msg number	*/
	int maxmsg;	/*max message number	*/
	int curmsg;	/*current number of msg	*/
	MSGINF msg[1];	/*message DYNAMIC array	*/
			/*at leat ONE		*/
	}MSGTYP;
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to increase and return a new	*/
/*	nummsg, if lock is not available or	*/
/*	time doesn't allow, return 0		*/
/*						*/
/************************************************/
static u_long newmsgnum(MSGTYP *ipcmsg,TIMESPEC *timeout)

{
u_long msgnum;

msgnum=(u_long)0;
if (ipc_getshraccess((void *)ipcmsg,timeout)==true) {
  ipcmsg->msgnum%=999999;
  ipcmsg->msgnum++;
  msgnum=ipcmsg->msgnum;
  (void) ipc_dropshraccess((void *)ipcmsg);
  }
return msgnum;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a MSGUID from	*/
/*	a  MSGREF information			*/
/*						*/
/************************************************/
static void getmsguid(char *msgref,MSGUID *msguid)

{
msguid->pidref=(pid_t)0;
msguid->ctxref=(u_long)0;
if (sscanf(msgref,"%d:%lu",&msguid->pidref,&msguid->ctxref)!=2)
  (void) alert(0,"gesipc.c:getmsguid, Unable to convert MSGREF <%s>",msgref);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to compose a MSGREF from	*/
/*	a MSGUID information, allocated memory	*/
/*	is dynamic and MUST be released later	*/
/*	on.					*/
/*						*/
/************************************************/
static char *mkmsgref(MSGUID *msguid)

{
char strloc[50];

(void) snprintf(strloc,sizeof(strloc),"%d:%lu",msguid->pidref,msguid->ctxref);
return strdup(strloc);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract data contents and	*/
/*	remove message with a specific message	*/
/*	number from the shared area.		*/
/*						*/
/************************************************/
static char *extract(MSGTYP *ipcmsg,u_long msgnum,struct timespec *timeout)

{
char *got;
int toremove;
MSGINF *info;

got=(char *)0;
info=(MSGINF *)0;
toremove=0;
if (ipc_getshraccess(ipcmsg,timeout)==true) {
  int gotit;
  int ndx;
  int i;

  ndx=0;
  gotit=false;
  for (i=0;i<ipcmsg->curmsg;i++) {
    if (ipcmsg->msg[i].msgnum==msgnum) {
      toremove++;
      ndx=i;
      }
    else {
      if (toremove>0)
        break;
      }
    }
  if (toremove>0) {
    MSGINF *ptr;
    int start;
    int tomove;

    info=(MSGINF *)calloc(toremove,sizeof(MSGINF));
    start=ndx-(toremove-1);
    ptr=&(ipcmsg->msg[start]);
    tomove=(ipcmsg->curmsg-start)-toremove;
    (void) memcpy(info,ptr,toremove*sizeof(MSGINF));
    /*removing from shared memory	*/
    (void) memcpy(ptr,ptr+toremove,tomove*sizeof(MSGINF));
    ipcmsg->curmsg-=toremove;
    }
  (void) ipc_dropshraccess(ipcmsg);
  }
if (info!=(MSGINF *)0) {
  int i;

  got=(char *)calloc(toremove,MINF);
  for (i=0;i<toremove;i++) {
    (void) memcpy(got+(MINF*i),info[i].info,MINF);
    }
  (void) free(info);
  }
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a message number 	*/
/*	addressed to the current process	*/
/*						*/
/************************************************/
static u_long getmsgnum(MSGTYP *ipcmsg,pid_t curpid,
			MSGUID *msguid,TIMESPEC *timeout)

{
u_long msgnum;

msgnum=(u_long)0;
if ((ipcmsg->curmsg>0)&&(ipc_getshraccess(ipcmsg,timeout)==true)) {
  int i;
  
  i=0;
  while (i<ipcmsg->curmsg) {
    MSGINF *msgloc;

    msgloc=ipcmsg->msg+i;
    if (msgloc->rcpt==curpid) {
      if ((msguid->ctxref==0)||(msguid->ctxref==msgloc->ctxref)) {
        msgnum=msgloc->msgnum;
	msguid->pidref=msgloc->pidref;
	msguid->ctxref=msgloc->ctxref;
        break;
	}
      }
    i++;
    }
  (void) ipc_dropshraccess(ipcmsg);
  }
return msgnum;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to add a message to the 	*/
/*	message queue.				*/
/*	If the message can't be stored, errno is*/
/*	set to ENOMEM and a 0 status is 	*/
/*	returned.				*/
/*	if everything is right, procedure 	*/
/*	return a non zero value			*/
/*						*/
/************************************************/
static u_long sendmsg(MSGTYP *ipcmsg,MSGUID *msguid,pid_t destination,
		      char *msg,TIMESPEC *timeout)

{
#define	OPIC	"gesipc.c:sendmsg,"

#define	MXSTRICK 300

int status;
pid_t curpid;
time_t curtime;
int taille;
int needed;
u_long msgnum;
pid_t pidref;
u_long ctxref;
MSGINF *msginf;
int i;

status=0;
curpid=getpid();
if ((msgnum=newmsgnum(ipcmsg,timeout))==0) {
  errno=EAGAIN;
  return status;
  }
if (msguid!=(MSGUID *)0) {
  ctxref=msguid->ctxref;
  pidref=msguid->pidref;
  }
else {
  ctxref=msgnum;
  pidref=curpid;
  }
curtime=rou_systime();
taille=strlen(msg);
needed=(taille/MINF)+1;
msginf=(MSGINF *)calloc(needed,sizeof(MSGINF));
(void) tracing(TRALOG,10,"To/From '%05d/%05d' SEND IPC "
			 "[%05d:%lu:%lu] msg <%s>",
			  destination,getpid(),pidref,ctxref,msgnum,msg);
for (i=0;i<needed;i++) {
  int max;

  max=taille-(MINF*i);
  if (max>MINF)
    max=MINF;
  msginf[i].stamp=curtime;
  msginf[i].rcpt=destination;
  msginf[i].from=curpid;
  msginf[i].pidref=pidref;
  msginf[i].ctxref=ctxref;
  msginf[i].msgnum=msgnum;
  (void) memcpy(msginf[i].info,msg+(MINF*i),max);
  }
errno=ENOMEM;
for (i=0;i<MXSTRICK;i++) {
  if ((ipcmsg->maxmsg-ipcmsg->curmsg)>(needed+10)) 
    break;
  (void) usleep(1000);
  }
if ((ipcmsg->maxmsg-ipcmsg->curmsg)>(needed+10))  {
  MSGINF *ptr;

  if (ipc_getshraccess((void *)ipcmsg,timeout)==true) {
    ptr=&(ipcmsg->msg[ipcmsg->curmsg]);
    (void) memcpy(ptr,msginf,needed*sizeof(MSGINF));
    ipcmsg->curmsg+=needed; 
    (void) ipc_dropshraccess((void *)ipcmsg);
    }
  if (kill(destination,SIGALRM)<0) {
    (void) alert(0,"%s NO kill to '%05d' (errno=<%s>)",
		    OPIC,destination,strerror(errno));
    }
  else
    status=msgnum;
  }
else {
  (void) alert(0,"%s Unable to post message needed='%d', current queue='%d'",
		 OPIC,needed,ipcmsg->curmsg);
  }
(void) free(msginf);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get and subtract a message	*/
/*	from the shared message list.		*/
/*	return a string stored within the 	*/
/*	message area, return the number of read	*/
/*	character, -1 if trouble		*/
/*	If no message in the waiting queue  	*/
/*	usleep for 'waiting' microsec		*/
/*						*/
/************************************************/
static char *getmsg(MSGTYP *ipcmsg,MSGUID *msguid,TIMESPEC *timeout)

{
char *got;
pid_t curpid;
u_long msgnum;

got=(char *)0;
curpid=getpid();
if ((msgnum=getmsgnum(ipcmsg,curpid,msguid,timeout))!=(u_long)0) {
  got=extract(ipcmsg,msgnum,timeout);
  (void) tracing(TRALOG,10,"By '%05d' GOT IPC [%05d:%lu:%lu] msg <%s>",
			    curpid,msguid->pidref,msguid->ctxref,
			    msgnum,got);
  }
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to assign a shared memory	*/
/*	area used to exchange message between	*/
/*	2 mentor process.			*/
/*						*/
/************************************************/
BOXTYP *ipc_getbox(int boxsize)

{
MSGTYP *box;

box=(MSGTYP *)ipc_getshrmem(boxsize+1,sizeof(MSGINF));
box->maxmsg=boxsize;
return (BOXTYP *)box;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free  a shared memory	*/
/*	area.					*/
/*						*/
/************************************************/
BOXTYP *ipc_dropbox(BOXTYP *box)

{
MSGTYP *msg;

msg=(MSGTYP *)box;
if (msg!=(MSGTYP *)0) {
  msg=(MSGTYP *)ipc_dropshrmem((void *)msg);
  }
return (BOXTYP *)msg;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove 'forgetted' message	*/
/*	from the message queue.			*/
/*						*/
/************************************************/
void ipc_cleanbox(BOXTYP *box)

{
static time_t lastcheck=(time_t)0;

time_t curtime;

curtime=rou_systime();
if ((lastcheck+15)<curtime) {
  MSGTYP *ipcmsg;
  int i;

  ipcmsg=(MSGTYP *)box;
  for (i=0;i<ipcmsg->curmsg;i++) {
    TIMESPEC timeout;

    timeout.tv_nsec=50000;	/*50 millisec	*/
    timeout.tv_sec=0;
    if (ipcmsg->msg[i].stamp>=lastcheck)
      continue;
    if (kill(ipcmsg->msg[i].rcpt,SIGCHECK)<0) {
      char *msg;

      msg=extract(ipcmsg,ipcmsg->msg[i].msgnum,&timeout);
      (void) alert(0,"ipc_cleanbox removed msg <%s> to '%05d'",
		      msg,ipcmsg->msg[0].rcpt);
      msg=rou_freestr(msg);
      i--;
      continue;	/*lets sync on next messages	*/
      }
    if ((ipcmsg->msg[i].stamp+60)<lastcheck) {
      MSGINF *ptr;

      ptr=&(ipcmsg->msg[i]);
      (void) alert(0,"ipc_cleanbox msg pill-up to/from "
		     "'%05d/%05d' (%03d/<%s>)",
		     ptr->rcpt,ptr->from,ptr->msgnum,ptr->info);
      }
    }
  lastcheck=curtime;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to post a message to a 	*/
/*	specific process.			*/
/*	If the message can't be sent, function	*/
/*	return a NULL pointer as message	*/
/*	reference.				*/
/*						*/
/************************************************/
char *ipc_postmsg(BOXTYP *box,pid_t destination,char *msg)

{
char *msgref;
MSGTYP *ipcmsg;
u_long msgnum;
TIMESPEC timeout;

msgref=(char *)0;
ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=40000;	/*40 msec at max to post*/
timeout.tv_sec=0;	/*a message in IPC	*/
if ((msgnum=sendmsg(ipcmsg,(MSGUID *)0,destination,msg,&timeout))!=0) {
  MSGUID msguid;

  msguid.pidref=getpid();
  msguid.ctxref=msgnum;
  msgref=mkmsgref(&msguid);
  }
return msgref;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get the first message 	*/
/*	for the current procees available	*/
/*	in the queue.				*/
/*						*/
/************************************************/
char *ipc_getpost(BOXTYP *box,char **msgref)

{
char *msg;
MSGTYP *ipcmsg;
TIMESPEC timeout;
MSGUID msguid;

msg=(char *)0;
ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=100000;	/*100 msec at max to get*/
timeout.tv_sec=0;	/*a message from IPC	*/
msguid.ctxref=(u_long)0;
if ((msg=getmsg(ipcmsg,&msguid,&timeout))!=(char *)0) {
  *msgref=mkmsgref(&msguid);
  } 
return msg;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get a specific reply to	*/
/*	a previously sent message.		*/
/*						*/
/************************************************/
char *ipc_getreply(BOXTYP *box,char *msgref)

{
MSGTYP *ipcmsg;
TIMESPEC timeout;
MSGUID msguid;

ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=80000;	/*80 msec at max to get	*/
timeout.tv_sec=0;	/*a message reply 	*/
(void) getmsguid(msgref,&msguid);
return getmsg(ipcmsg,&msguid,&timeout);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to post a reply to a 		*/
/*	previously get message.			*/
/*	Return zero if trouble otherwise the	*/
/*	message number.				*/
/*						*/
/************************************************/
int ipc_postreply(BOXTYP *box,char *msgref,char *msg)

{
MSGTYP *ipcmsg;
TIMESPEC timeout;
MSGUID msguid;

ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=100000; /*100 msec at max to get*/
timeout.tv_sec=0;	/*a message reply 	*/
(void) getmsguid(msgref,&msguid);
return sendmsg(ipcmsg,&msguid,msguid.pidref,msg,&timeout);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to relay message to another	*/
/*	process while keep the message origin.	*/
/*	Return zero if trouble otherwise the	*/
/*	message number.				*/
/*						*/
/************************************************/
int ipc_relaymsg(BOXTYP *box,char *msgref,pid_t destination,char *msg)

{
MSGTYP *ipcmsg;
TIMESPEC timeout;
MSGUID msguid;

ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=150000;	/*150 msec at max to get*/
timeout.tv_sec=0;	/*a message reply 	*/
(void) getmsguid(msgref,&msguid);
return sendmsg(ipcmsg,&msguid,destination,msg,&timeout);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get one UNIQUE message	*/
/*	reference.				*/
/*						*/
/************************************************/
char *ipc_getonemsgref(BOXTYP *box)

{
char *msgref;
u_long msgnum;
MSGTYP *ipcmsg;
TIMESPEC timeout;

msgref=(char *)0;
ipcmsg=(MSGTYP *)box;
timeout.tv_nsec=15000;	/*15 msec at max to get	*/
timeout.tv_sec=0;	/*message number	*/
if ((msgnum=newmsgnum(ipcmsg,&timeout))!=(u_long)0) {
  MSGUID msguid;

  msguid.pidref=getpid();
  msguid.ctxref=msgnum;
  msgref=mkmsgref(&msguid);
  }
return msgref;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove ALL msg addressed	*/
/*	to a specific process.			*/
/*						*/
/************************************************/
void ipc_purgemsgs(BOXTYP *box,pid_t rcpt)

{
MSGTYP *ipcmsg;
int i;

ipcmsg=(MSGTYP *)box;
for (i=0;i<ipcmsg->curmsg;i++) {
  if (ipcmsg->msg[i].rcpt==rcpt) {
    TIMESPEC timeout;
    char *msg;

    timeout.tv_nsec=50000;	/*50 millisec	*/
    timeout.tv_sec=0;
    msg=extract(ipcmsg,ipcmsg->msg[i].msgnum,&timeout);
    (void) alert(0,"ipc_puregmsgs removed msg <%s> to '%05d'",
		    msg,ipcmsg->msg[0].rcpt);
    msg=rou_freestr(msg);
    i--;
    continue;	/*lets sync on next messages	*/
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to wait for a specific time	*/
/*	or until a message is ready, whichever	*/
/*	first. Return true is message available	*/
/*						*/
/************************************************/
int ipc_waitpost(BOXTYP *box,TIMESPEC *timeout)

{
int found;
MSGTYP *ipcmsg;
pid_t curpid;

found=false;
ipcmsg=(MSGTYP *)box;
curpid=getpid();
while (found==false) {
  if ((ipcmsg->curmsg>0)&&(ipc_getshraccess(ipcmsg,timeout)==true)) {
    register int i;
    register MSGINF *msgloc;

    i=0; 
    msgloc=ipcmsg->msg;
    while (i<ipcmsg->curmsg) {
      if (msgloc->rcpt==curpid) {
	found=true;
	break;
	}
      msgloc++;
      i++;
      }
    (void) ipc_dropshraccess((void *)ipcmsg);
    }
  if ((found==true)||(quickout==true))
    break;
  if (nanosleep(timeout,timeout)<0) {
    switch (errno) {
      case EINTR	:
	break;
      default		:
	(void) alert(0,"gesipc.c:ipc_waitpost, unexpected error (error=<%s>)",
			strerror(errno));
	break;
      }
    }
  else  {
    timeout->tv_nsec=0;
    timeout->tv_sec=0;
    }
  if ((timeout->tv_nsec==(long)0)&&(timeout->tv_sec==0))
    break;
  }
return found;
}
